"""
Contains class representation of executable code block.
"""
from prompt_toolkit import prompt
from prompt_toolkit.styles import style_from_pygments
from pygments import highlight
from pygments.formatters import Terminal256Formatter
from pygments.lexers import get_lexer_by_name
import logging
import subprocess
import sys

class DocCode(object):
    """Single multi-line code block executed as a script.

    Attributes:
        interpreter (str): Interpreter used to run the code.
        code (str): Base code loaded during initialization.
        user_code (str): User modified version of the code (will be used
            instead if is not None or empty string).
        process (subprocess.Popen): Process object running the interpreter.
        output (dict): Dictinary containing 'stdout' and 'retcode'.
    """
    def __init__(self, code, interpreter, darkbg=True):
        if darkbg:
            from pygments.styles.monokai import MonokaiStyle as HighlightStyle
            self.HighlightStyle = HighlightStyle
        else:
            from pygments.styles.manni import ManniStyle as HighlightStyle
            self.HighlightStyle = HighlightStyle
        self.interpreter = interpreter
        self.code = code
        self.user_code = ''
        self.process = None
        self.output = { 'stdout':'', 'retcode':None }

    def get_lexer_class(self):
        lexer_class = None
        try:
            # try because lexer may not exist for current interpreter
            return get_lexer_by_name(self.interpreter).__class__
        except:
            # no lexer, return plain text
            return None

    def __str__(self):
        lexer_class = self.get_lexer_class()
        code = self.user_code.strip() or self.code
        if lexer_class:
            return highlight(
                code,
                lexer_class(),
                Terminal256Formatter(style=self.HighlightStyle)
                )
        return code

    def get_dict(self):
        return {
            'interpreter': self.interpreter,
            'code': self.code,
            'user_code': self.user_code,
            'output': self.output,
        }

    def prompt_user(self, prompt_text='» '):
        self.user_code = prompt(
            prompt_text,
            default = self.code,
            lexer = self.get_lexer_class(),
            style = style_from_pygments(self.HighlightStyle)
            )

    def print_stdout(self):
        assert self.process
        line = self.process.stdout.readline().decode('utf-8')
        self.output['stdout'] += line
        print(line, end='')

    def is_running(self):
        return self.process and self.process.poll() is None

    def run(self):
        if not self.process:
            code = self.user_code.strip() or self.code
            logging.debug('Running code {}'.format(code))
            self.process = subprocess.Popen(
                [self.interpreter, '-c', code],
                stdout=subprocess.PIPE,
                stderr=sys.stdout.buffer,
                shell=False,
                )
        while self.is_running():
            self.print_stdout()
        self.output['retcode'] = self.process.poll()

    def kill(self):
        if self.process:
             self.process.kill()





"""
Classes and tools for parsing markdown docs and manipulating their execution.
"""
from bs4 import BeautifulSoup
from collections import OrderedDict
from prompt_toolkit import prompt
from rundoc import BadEnv
from rundoc.doc_code import DocCode
from time import sleep
import json
import markdown
import os
import re
import sys

class clr:
    ''' 
    ANSI colors for pretty output.
    '''
    red = '\033[91m'
    green = '\033[92m'
    blue = '\033[94m'
    yellow = '\033[93m'
    bold = '\033[1m'
    underline = '\033[4m'
    end = '\033[0m'

class OrderedEnv(OrderedDict):
    """Dictionary of environment variables.

    Preserves order of variables as found in string. Lets you load a string of
    variable definitions (new line separated var=value pairs).
    You can append (variable, value) pairs and extend with objects of the same
    type.
    Can prompt user to modify variables. Let's you load variables into os.
    Functions append, extend and import_string contain  'collect_existing_env'
    argument. If you set it to True, then these functions will not overwrite
    existing and os variables in case you set them to empty values (assumes
    that giving empty string as value means you don't with to change those
    variables).
    """
    def __init__(self, title=None):
        super().__init__()
        self.title = title

    def __str__(self):
        return "\n".join([ var+"="+self[var] for var in self ])

    def append(self, var, val, collect_existing_env=True):
        if collect_existing_env:
            val = val or self.get(var) or os.environ.get(var, '')
        self[var] = val

    def extend(self, env, collect_existing_env=True):
        for var in env:
            self.append(var, env[var], collect_existing_env)

    def import_string(self, env_string, collect_existing_env=True):
        for line in env_string.strip().split('\n'):
            if not line: continue
            if '=' not in line:
                raise BadEnv("Bad environment line: {}".format(line))
            var, val = line.split('=', 1)
            var = var.strip()
            val = val.strip()
            if not var:
                raise BadEnv("Bad environment line: {}".format(line))
            self.append(var, val, collect_existing_env)

    def prompt(self):
        if not len(self):
            return
        print(self.title)
        msg = "{}\tConfirm/supply/modify environment variables."
        msg += "\n\tPress Return to finish.{}"
        msg = msg.format(clr.bold, clr.end)
        print(msg)
        env_string = str(self)
        env_string = prompt( '» ', default = env_string )
        self.clear()
        self.import_string(env_string, collect_existing_env=False)

    def prompt_missing(self):
        missing = self.__class__(self.title)
        for var in self:
            if not self[var]:
                missing.append(var, '')
        missing.prompt()
        self.extend(missing, collect_existing_env=False)

    def load(self):
        for var in self:
            os.environ[var] = self[var]


class DocCommander(object):
    """
    Manages environment and DocCode objects and executes them in succession.
    """
    def __init__(self):
        self.env = OrderedEnv(
            "\n{}==== env variables{}".format(clr.bold, clr.end)
            )
        self.secrets = OrderedEnv(
            "\n{}==== secrets{}".format(clr.bold, clr.end)
            )
        self.doc_codes = []
        self.running = False
        self.failed = False
        self.current_doc_code = None
        self.current_step = None

    def get_dict(self):
        return {
            'env': self.env,
            'code_blocks': [ x.get_dict() for x in self.doc_codes ]
        }

    def add(self, code, interpreter='bash', darkbg=True):
        if not interpreter:
            interpreter = 'bash'
        self.doc_codes.append(DocCode(code, interpreter, darkbg))

    def die_with_grace(self):
        if self.running:
            self.current_doc_code.kill()
            print("\n==== {}Quit at step {} with keyboard interrupt.{}\n".format(
                clr.red,
                self.current_step,
                clr.end,
                )
            )

    def run(self, step=1, yes=False, pause=0):
        """Run all the doc_codes one by one starting from `step`.

        Args:
            step (int): Number of step to start with. Defaults to 1. Steps
                start at 1.
            yes (bool): Auto-confirm all steps without user interaction.
                Defaults to False.
            pause (int): Add a delay in seconds before the start of each step.
                Makes sense only when 'yes' is set to True. Defaults to 0.

        Returns:
            JSON representation of code blocks and outputs.
        """
        assert self.running == False
        assert self.failed == False
        if yes:
            self.env.prompt_missing()
            self.secrets.prompt_missing()
        else:
            self.env.prompt()
            self.secrets.prompt()
        msg = "\n{}Running code blocks from supplied documentation."
        msg += "\nModify and/or confirm displayed code by pressing Return.{}"
        print(msg.format(clr.bold, clr.end))
        self.env.load()
        self.secrets.load()
        self.running = True
        for doc_code in self.doc_codes[step-1:]:
            self.current_step = step
            self.current_doc_code = doc_code
            prompt_text = "\n{}=== Step {} [{}]{}".format(
                clr.bold,
                step,
                doc_code.interpreter,
                clr.end,
                )
            print(prompt_text)
            if yes:
                print(doc_code)
                sleep(pause)
            else:
                doc_code.prompt_user()
            self.current_doc_code.run() # run in blocking manner
            if doc_code.output['retcode'] != 0:
                self.failed = True
            if self.failed:
                self.running = False
                print("==== {}Failed at step {} with exit code '{}'{}\n".format(
                    clr.red,
                    step,
                    doc_code.output['retcode'],
                    clr.end,
                    )
                )
                self.current_doc_code = None
                break
            self.current_doc_code = None
            print("{}==== Step {} done{}\n".format(clr.green, step, clr.end))
            step += 1
        return json.dumps(self.get_dict(), sort_keys=True, indent=4)

def parse_doc(mkd_file_path, tags="", darkbg=True):
    """Parse code blocks from markdown file and return DocCommander object.

    Args:
        mkd_file_path (str): Path to markdown file.
        tags (str): Code highlight specifier in markdown. We can use this to
            filter only certain code blocks. If it's set to empty string or
            None, all code blocks will be used. Defaults to "bash".

    Returns:
        DocCommander object.
    """
    mkd_data = ""
    with open(mkd_file_path, 'r') as f:
        mkd_data = f.read()
    html_data = markdown.markdown(
        mkd_data,
        extensions=['toc', 'tables', 'footnotes', 'fenced_code']
        )
    soup = BeautifulSoup(html_data, 'html.parser')
    # collect all elements with selected tags as classes
    classes = re.compile(
        "(^|_)({})(_|$)".format('|'.join(tags.split(','))) if tags else '^(?!(env|secret)).*'
        )
    code_block_elements = soup.findAll('code', attrs={"class":classes,})
    commander = DocCommander()
    for element in code_block_elements:
        class_name = element.get_attribute_list('class')[0]
        interpreter = None
        if class_name:
            interpreter = class_name.split("_")[0]
        commander.add(
            element.getText(),
            interpreter,
            darkbg
        )
    # get env blocks
    classes = re.compile("^env(iron(ment)?)?$")
    env_elements = soup.findAll('code', attrs={"class":classes,})
    env_string = "\n".join([ x.string for x in env_elements ])
    commander.env.import_string(env_string)
    # get secrets blocks
    classes = re.compile("^secrets?$")
    secrets_elements = soup.findAll('code', attrs={"class":classes,})
    secrets_string = "\n".join([ x.string for x in secrets_elements ])
    commander.secrets.import_string(secrets_string)
    return commander

def parse_output(output_file_path):
    """Load json output, create and return DocCommander object.

    Args:
        output_file_path (str): Path to saved output file.

    Returns:
        DocCommander object.
    """
    output_data = None
    with open(output_file_path, 'r') as f:
        output_data = f.read()
    data = json.loads(output_data)
    commander = DocCommander()
    for d in data:
        doc_code = DocCode(d['code'], d['interpreter'])
        doc_code.user_command = d['user_code']
        commander.doc_codes.append(doc_code)
    return commander




"""
Main module for rundoc command line utility.
"""
from rundoc import BadEnv
from rundoc.doc_commander import parse_doc, parse_output 
import argcomplete
import argparse
import logging
import rundoc
import sys

def __parse_args():
    """Parse command line arguments and return an argparse object."""
    parser = argparse.ArgumentParser(
        prog = "rundoc",
        description="Run code from markdown code blocks in controlled manner",
        )
    parser.add_argument(
        "-v", "--version", action="store_true",
        help="Show version info and exit."
        )
    parser.add_argument(
        "-d", "--debug", action="store_true",
        help="Enable debug mode with output of each action in the log."
        )
    subparsers = parser.add_subparsers(
        description="(use each command with -h for more help)",
        dest="cmd",
        )

    parser_run = subparsers.add_parser(
        "run",
        description="Run markdown as a script."
        )
    parser_run.add_argument(
        "mkd_file_path", type=str, action="store",
        help="Markdown file path."
        )
    parser_run.add_argument(
        "-t", "--tags", action="store",
        help='''Coma-separated list of tags (e.g. -t
            bash,bash_proj-2,python3_proj2). Part of tag until first underscore
            will be used as selected interpreter for that code. If no tags are
            provided, all code blocks will be used. Untagged code blocks will
            use bash as default interpreter.'''
        )
    parser_run.add_argument(
        "-p", "--pause", type=int, action="store",
        help='''Used in combination with -y. Pause N seconds before each code
                block. Defaults to 0.'''
        )
    parser_run.add_argument(
        "-s", "--step", type=int, action="store",
        help="Start at step STEP. Defaults to 1."
        )
    parser_run.add_argument(
        "-o", "--output", type=str, action="store",
        help='''Output file. All codes, modifications and output of execution
            will be saved here. This file can be later used as an argument to
            'rerun' command which will execute all steps without any user
            interaction.'''
        )
    parser_run.add_argument(
        "--light", action="store_true",
        help="Light terminal background."
        )
    parser_run.add_argument(
        "-y", "--yes", action="store_true",
        help="Confirm all steps with no user interaction."
        )
    parser_run.set_defaults(
        tags="",
        pause=0,
        step=1,
        output=None,
        )

    parser_rerun = subparsers.add_parser(
        "rerun",
        description="Run codes saved as json output by 'run' command."
        )
    parser_rerun.add_argument(
        "saved_output_path", type=str, action="store",
        help="Path of saved output file from 'run' or 'rerun' command."
        )
    parser_rerun.add_argument(
        "-s", "--step", type=int, action="store",
        help="Start at step STEP. Defaults to 1."
        )
    parser_rerun.add_argument(
        "-o", "--output", type=str, action="store",
        help='''Output file. All codes, modifications and output of execution
            will be saved here. This file can be later used as an input
            argument to 'rerun' command.'''
        )
    parser_rerun.set_defaults(
        step=1,
        output=None,
        )

    argcomplete.autocomplete(parser)
    return parser.parse_args()

def main():
    args = __parse_args()
    logging.basicConfig(
        format = '%(asctime)s.%(msecs)03d, %(levelname)s: %(message)s',
        datefmt = '%Y-%m-%d %H:%M:%S',
        filename = None,
        level = logging.DEBUG if args.debug else logging.CRITICAL,
        )
    if args.version:
        print("rundoc {} - Copyright {} {} <{}>".format(
            rundoc.__version__,
            rundoc.__year__,
            rundoc.__author__,
            rundoc.__author_email__,
            ))
        sys.exit(0)
    output = ""

    if args.cmd == 'run':
        try:
            darkbg = not args.light
            commander = parse_doc(args.mkd_file_path, args.tags, darkbg=darkbg)
        except BadEnv as e:
            print("{}{}{}".format(clr.red, e, clr.end))
            sys.exit(1)
        try:
            output = commander.run(
                step=args.step, yes=args.yes, pause=args.pause)
        except KeyboardInterrupt:
            commander.die_with_grace()
        except BadEnv as e:
            print("{}{}{}".format(clr.red, e, clr.end))
            sys.exit(1)
    if args.cmd == 'rerun':
        commander = parse_output(args.saved_output_path)
        try:
            output = commander.run(step=args.step, yes=True)
        except KeyboardInterrupt:
            commander.die_with_grace()
        except BadEnv as e:
            print("{}{}{}".format(clr.red, e, clr.end))
            sys.exit(1)
    if args.cmd in ('rerun', 'run') and args.output:
        with open(args.output, 'w+') as f:
            f.write(output)

if __name__ == '__main__':
    main()




"""
Tool that let's you run code blocks from a markdown file in controlled manner.
"""

__version__ = "0.1.11"
__licence__ = "BSD"
__year__ = "2017"
__author__ = "EclecticIQ"
__author_email__ = "info@eclecticiq.com"


class RundocException(Exception):
    """Generic rundoc exception."""
    pass

class BadEnv(RundocException):
    """Attempt to load invalid environment name/value"""
    pass




#!/usr/bin/env python3

from setuptools import setup, find_packages
import rundoc

setup(
    name = 'rundoc',
    description = rundoc.__doc__.strip(),
    url = 'https://github.com/EclecticIQ/rundoc',
    download_url = 'https://github.com/EclecticIQ/rundoc/archive/'+rundoc.__version__+'.tar.gz',
    version = rundoc.__version__,
    author = rundoc.__author__,
    author_email = rundoc.__author_email__,
    license = rundoc.__licence__,
    packages = [ 'rundoc' ],
    entry_points={ 
        'console_scripts': [
            'rundoc=rundoc.__main__:main',
        ],
    },
    install_requires = [
        'markdown>=2.6.9',
        'argcomplete>=1.9.2',
        'bs4',
        'pygments>=2.2.0',
        'prompt_toolkit>=1.0.15',
    ],
    python_requires=">=3.4.7",
)

